/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: LiveDataServer						*/
/* Module Description: see LiveDataServer.h for more information	*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Matthew Evans  (626)395-8507	  -             mevans@ligo.caltech.edu	*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#include "LiveDataServer.hh"
#include <iostream>

// define DEBUG
#define NDS_DEBUG 0
#define LCL_DEBUG 0
#ifdef DEBUG
#undef DEBUG
#endif
#if LCL_DEBUG
#define DEBUG(x) cerr << "DEBUG: LDS - " << x
#else
#define DEBUG(x)
#endif

// define ERROR
#ifdef ERROR
#undef ERROR
#endif
#define ERROR(x) cout << "ERROR: LDS - " << x

// swap bytes?
#define BYTE_SWAP 1
#define SWAP_BYTES(a, b) a^=b; b^=a; a^=b

// ----------------------------------------------------------------
// PUBLIC
// ----------------------------------------------------------------

bool LiveDataServer::initNDS(const char* ndsName)
{
  // clean and initialize
  clean(NDS_SOCK);
  init(NDS_SOCK);

  // Open DAQ Socket
  if( !openNDS(ndsName) && !openNDS(getenv("LIGONDSIP")) &&
      !openNDS("fb0") && !openNDS("fb1") )
  {
    ERROR("Unable to open socket to NDS." << endl);
    return false;				// ======== RETURN
  }

  // Get number of channels
  DEBUG("Getting available channel count from NDS..." << endl);
  numAvl = nds.Available(NULL, 0);
  if( numAvl < 0 )
  {
    numAvl = 0;
    ERROR("Unable to get channel list from NDS." << endl);
    return false;				// ======== RETURN
  }

  // Get channel list
  DEBUG("Number of channels returned is " << numAvl << endl);
  avlList = new DAQDChannel[numAvl];
  if( numAvl != nds.Available(avlList, numAvl) )
  {
    numAvl = 0;
    ERROR("Inconsistent NDS channel list." << endl);
    return false;				// ======== RETURN
  }

  // Set NDS debug level (done here to avoid channel listing spewage)
  nds.setDebug(NDS_DEBUG);

  return true;
}

const LiveData* LiveDataServer::initChannels
 (const char** chanName, int numChanArg)
{
  int j;				// index used to find channels
  int offNext = 0;			// offset of next data channel
  int len = 0;				// length of this data channel

  // clean and initialize
  if( avlList == NULL )
  {
    ERROR("NDS not ready, call initNDS." << endl);
    return NULL;
  }
  clean(CHANNELS);
  init(CHANNELS);

  // allocate LiveData array
  numChan = numChanArg;
  chanList = new LiveData[numChan];

  // ---- Add channels
  for( int i = 0; i < numChan; ++i )
  {
    // find channel in avlList
    j = 0;
    while( j < numAvl  && strcasecmp(avlList[j].mName, chanName[i]) )
      ++j;

    if( j == numAvl )
    {
      ERROR("Invalid channel name " << chanName[i] << endl);
      return NULL;				// ======== RETURN
    }

    // check mBPS
    switch( avlList[j].mDatatype )
    {
      case SHORT: checkBPS(avlList + j, sizeof(short)); break;
      case INT: checkBPS(avlList + j, sizeof(int)); break;
      case LONG: checkBPS(avlList + j, sizeof(long)); break;
      case FLOAT: checkBPS(avlList + j, sizeof(float)); break;
      case DOUBLE: checkBPS(avlList + j, sizeof(double)); break;
      case QUAD: checkBPS(avlList + j, sizeof(long double)); break;
    }

    // add to NDS
    chanList[i].daq = avlList[j];
    nds.AddChannel(chanList[i].daq);
  }

  // ---- Set offsets and data lengths
  // iterate through channels in NDS channel list
  for( DAQSocket::Channel_iter i = nds.mChannel.begin();
       i != nds.mChannel.end(); ++i )
  {
    // compute data length
    len = i->second.mRate / BLOCKS_PER_SECOND;

    // find matching channel (or channels)
    for( int j = 0; j < numChan; ++j )
      if( i->first == chanName[j] )
      {
        chanList[j].off = offNext;
        chanList[j].len = len;
        chanList[j].data = new double[len];
        DEBUG(chanName[j]
              << ", offset " << offNext
              << ", length " << len
              << ", rate " << chanList[j].daq.mRate
              << ", type " << chanList[j].daq.mDatatype
              << ", BPS "  << chanList[j].daq.mBPS << endl);
      }
      
    // prepare for next step, prevent erroneous offsets
    if( i->second.mBPS < 100 )
      offNext += len * i->second.mBPS;
    else
    {
       ERROR("Invalid BPS of " << i->second.mBPS
             << " given for channel " << i->second.mName << endl);
       return false;
    }
  }

  // set final data length and allocate data buffer
  dataLen = offNext;
  if( dataLen > 16 * numChan )
    bufLen = sizeof(DAQDRecHdr) + dataLen;
  else
    bufLen = sizeof(DAQDRecHdr) + numChan * 16;
  buf = new char[bufLen];
  DEBUG("Expected data length is " << dataLen << " bytes" << endl);
  DEBUG("Buffer length is " << bufLen << " bytes" << endl);

  return chanList;
}

bool LiveDataServer::initDataFlow()
{
  const DAQDRecHdr* hdr = (DAQDRecHdr*)(buf);
  int rtnVal;

  // clean and initialize
  if( buf == NULL )
  {
    ERROR("Channels not ready, call initChannels." << endl);
    return false;				// ======== RETURN
  }
  clean(SEQUENCE);
  init(SEQUENCE);

  // request NDS data flow
  rtnVal = nds.RequestOnlineData(true,-1);
  if( rtnVal )
  {
    ERROR("NDS server error #" << rtnVal << endl);
    return false;				// ======== RETURN
  }

  // get first sequence number
  for( int i = 0; i < 3 && seqNum == -1; ++i )
  {
    rtnVal = nds.GetData(buf, bufLen, -1);
    DEBUG("GetData returned " << rtnVal << " bytes of data"
          " with SeqNum " << hdr->SeqNum << endl);

    if( rtnVal == dataLen )
      seqNum = hdr->SeqNum;
    else if( rtnVal > 0 && hdr->SeqNum != -1 )
    {
      ERROR("Incorrect data length, expected " << dataLen
            << ", got " << rtnVal << endl);
      return false;				// ======== RETURN
    }
  }

  // check for error
  if( seqNum == -1 )
  {
    ERROR("Data Receiving Error " << rtnVal << endl);
    return false;				// ======== RETURN
  }

  // copy data into chanList buffers
  copyData();

  return true;
}

bool LiveDataServer::getData(bool ignoreSeqNum)
{
  const DAQDRecHdr* hdr = (DAQDRecHdr*)(buf);
  int rtnVal;

  // check seqNum
  if( seqNum == -1 )
  {
    ERROR("Data flow not ready, call initDataFlow." << endl);
    return false;				// ======== RETURN
  }

  // get data
  rtnVal = nds.GetData(buf, bufLen, -1);
  DEBUG("GetData returned " << rtnVal << " bytes of data"
        " with SeqNum " << hdr->SeqNum << endl);

  // check sequence number
  ++seqNum;
  if( hdr->SeqNum != seqNum && !ignoreSeqNum)
  {
    ERROR("Data Sequence Error, expected " << seqNum
           << ", got " << (hdr->SeqNum) << endl);
    return false;				// ======== RETURN
  }
  seqNum = hdr->SeqNum;

  // check data length
  if( rtnVal != dataLen )
  {
    ERROR("Incorrect data length, expected " << dataLen
          << ", got " << rtnVal << endl);
    return false;				// ======== RETURN
  }

  // copy data into chanList buffers
  copyData();
  return true;
}

// ----------------------------------------------------------------
// PROTECTED
// ----------------------------------------------------------------

bool LiveDataServer::openNDS(const char* ndsName)
{
  int rtnVal;

  if( ndsName == NULL )
    return false;

  DEBUG("Connecting to NDS \"" << ndsName << "\"..." << endl);
  rtnVal = nds.open(ndsName);
  if( rtnVal != 0 )
    DEBUG("Unable to connect to NDS \"" << ndsName
           << "\" error " << rtnVal << endl);

  return rtnVal == 0;
}

void LiveDataServer::checkBPS(DAQDChannel* chan, int size)
{
  if( chan->mBPS != size )
  {
    DEBUG("BPS mismatch for " << chan->mName
          << ", expected " << size
          << " for type " << chan->mDatatype
          << ", got " << chan->mBPS << endl);
    chan->mBPS = size;
  }
}

void LiveDataServer::copyData()
{
  char* data;
  double* chanData;

  // copy data into chanList buffers
  for( int i = 0; i < numChan; ++i )
  {
    chanList[i].seqNum = seqNum;		// set sequence number
    data = buf + sizeof(DAQDRecHdr) + chanList[i].off;
    chanData = chanList[i].data;		// set local pointer

    for( int j = 0; j < chanList[i].len; ++j )
    {
      // swap bytes, if necessary
      #if BYTE_SWAP
      switch( chanList[i].daq.mBPS )
      {
        case 2:					// 2 bytes per sample
	SWAP_BYTES(data[0], data[1]);
        break;
      
        case 4:					// 4 bytes per sample
	SWAP_BYTES(data[0], data[3]);
	SWAP_BYTES(data[1], data[2]);
        break;

        case 8:					// 8 bytes per sample
	SWAP_BYTES(data[0], data[7]);
	SWAP_BYTES(data[1], data[6]);
	SWAP_BYTES(data[2], data[5]);
	SWAP_BYTES(data[3], data[4]);

        default:
        for( int k = 0; k < chanList[i].daq.mBPS / 2 ; k++) {
          SWAP_BYTES(data[k], data[chanList[i].daq.mBPS - k - 1]); }
      }
      #endif

      // copy data to chanData
      switch( chanList[i].daq.mDatatype )
      {
        case SHORT: *chanData = *((short*)data); break;
        case INT: *chanData = *((int*)data); break;
        case LONG: *chanData = *((long*)data); break;
        case FLOAT: *chanData = *((float*)data); break;
        case DOUBLE: *chanData = *((double*)data); break;
        case QUAD: *chanData = *((long double*)data); break;
      }

      // move pointers
      data += chanList[i].daq.mBPS;
      ++chanData;
    }
  }
}

void LiveDataServer::init(ResetLevel lvl)
{
  switch( lvl )
  {
    case NDS_SOCK:
    numAvl = 0;
    avlList = NULL;

    case CHANNELS:
    numChan = 0;
    chanList = NULL;

    dataLen = 0;
    bufLen = 0;
    buf = NULL;

    case SEQUENCE:
    seqNum = -1;
  }
}

void LiveDataServer::clean(ResetLevel lvl)
{
  int rtnVal;

  switch( lvl )
  {
    case NDS_SOCK:
    DEBUG( "Closing NDS..." << endl; )
    nds.close();
    delete[] avlList;
    DEBUG( "Closed NDS. " << endl; )

    case CHANNELS:
    for( int i = 0; i < numChan; ++i )
      delete chanList[i].data;
    delete chanList;
    delete buf;

    case SEQUENCE:
    DEBUG( "Stoping data flow..." << endl; )
    rtnVal = nds.StopWriter();
    DEBUG( "Data flow stopped: " << rtnVal << endl; )
  }
}


